#!/usr/bin/perl
# 
# Copyright (C) 2015 Roland Laifer 
#                    Karlsruhe Institute of Technology (KIT), 
#                    Steinbuch Centre for Computing (SCC)
#                    76128 Karlsruhe, Germany
#                    Roland.Laifer (at) kit.edu
#
# ------------------------------------------------------------------------
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------
#
# $Log: stat_ost_dev.pl,v $
# Revision 1.2  2015/10/29 16:38:46  rz54
# Lots of changes to better conform to good perl coding style
#
# Revision 1.1  2015/10/29 16:05:42  rz54
# Initial revision
#
#
#
# $Id: stat_ost_dev.pl,v 1.2 2015/10/29 16:38:46 rz54 Exp rz54 $

use strict;
use warnings;

use Getopt::Std;
our($opt_D, $opt_f, $opt_h, $opt_v, $opt_b, $opt_d, $opt_e, $opt_p, $opt_o, $opt_r, $opt_s);

# get script name
(my $myname = $0)	=~ s#^.*/##;
# path to debugfs
my $debugfs = "/sbin/debugfs";


sub usage {
    print "Usage: $myname [-h] [-D] [-b number] -d name {-e number | -o file} [-p number] [-r number] [-s number] [-v]\n";
    print "$myname does efficient stat on many objects of Lustre OST device with debugfs.\n";
    print "This script was mainly used to check if Lustre quotas are wrong as a result\n";
    print "of the bug reported in https://jira.hpdd.intel.com/browse/LU-4345 .\n";
    print "For details check Roland Laifer's talk at LAD'15,\n";
    print "see http://www.scc.kit.edu/scc/docs/Lustre/kit_lad15_20150922.pdf\n";
    print " -b number     number of minimum object ID (default: 0)\n";
    print " -d name       device name of OST device\n";
    print " -e number     number of maximum object ID (default: 0)\n";
    print " -f            force to continue even if OST device is not mounted\n";
    print " -o file       file with object IDs to pipe at the same time to debugfs\n";
    print " -p number     number of object IDs to pipe at the same time to debugfs\n";
    print "               (default: 100)\n";
    print " -r number     number of object IDs to pipe to the same debugfs call\n";
    print "               (default: 10000)\n";
    print " -s number     number of seconds to sleep after option -p pipes (default: 0)\n";
    print " -h            this help\n";
    print " -D            debug\n";
    print " -v            verbose\n\n";
    print "Example: $myname -e 1000 p 1000 -r 10000 -d /dev/mapper/ost_lustre1_0\n";
    exit 0;
}


# get options
getopts("Dfhvb:d:e:p:o:r:s:") || usage;
usage if ( $opt_h );
usage unless ($#ARGV == -1);
usage unless ( $opt_d );
usage unless ( $opt_e || $opt_o );

# variables to store options or default values
my ($beginobj, $endobj, $statcount, $restartcount, $sec);

if ( $opt_b ) {
    $beginobj = $opt_b;
} else {
    $beginobj = 0;
}
if ( $opt_e ) {
    $endobj = $opt_e;
}
if ( $opt_p ) {
    $statcount = $opt_p;
} else {
    $statcount = 100;
}
if ( $opt_r ) {
    $restartcount = $opt_r;
} else {
    $restartcount = 10000;
}
if ( $opt_s ) {
    $sec = $opt_s;
} else {
    $sec = 0;
}

# Check device name 
my @lustrefs = `cat /proc/mounts | grep lustre | egrep "^$opt_d"`;
if ( $#lustrefs == -1 ) {
    print STDERR "Warning: Possibly wrong OST device name ($opt_d) in option -d!\n";
    if ( $opt_f) {
	print STDERR "Did not find OST device name in /proc/mounts. Exiting...\n";
	exit 1;
    } else {
	print STDERR "Continuing anyway because of option -f.\n";
    }
}

# Read objects from file into arry 
# array of objects in case they are read from a file
my @objects;
# index in case objects are read from a file
my $index;
if ( $opt_o ) {
    unless ( -e $opt_o ) {
	print STDERR "File $opt_o does not exist. Exiting...\n";
	exit 1;
    }
    @objects = `sort -n "$opt_o"`;
    $endobj = $objects[ $#objects ];
    print "Last object: $endobj\n" if $opt_D;
}

# Now do the main task and execute a number of stat commands 
# at the same time with debugfs
# object ID which is used for iteration
my $objid;
if ( $opt_o ) {
    $index = 0;
    $objid = $objects[$index];
} else {
    $objid = $beginobj;
}
my $rcount = 0;
# command to start debugfs
my $cmd;
if ( $opt_D) {
    $cmd = "$debugfs -c $opt_d";
} else {
    $cmd = "$debugfs -c $opt_d 2> /dev/null";
}
open my $dbgfs_h, "|-", $cmd or die "*** Command $cmd failed: $!\n";
# iterate over all object IDs 
while ( $objid <= $endobj ) {
    my $pcount = 0;
    my $statcmds = "";
    # prepare number of commands to execute at the same time
    while ( $pcount < $statcount && $objid <= $endobj ) {
	$statcmds .= sprintf("stat /O/0/d%d/%d\n",$objid % 32,$objid);
	if ( $opt_o ) {
	    unless ( $objid == $endobj ) {
		$index++;
		$objid = $objects[$index];
	    } else {
		$objid++;
	    }
	} else {
	    $objid++;
	}
	$pcount++;
	$rcount++;
    }
    print "Stat commands:\n$statcmds" if $opt_D;
    print $dbgfs_h "$statcmds\n";
    # possibly restart debugfs 
    if ( $rcount >= $restartcount) {
	print STDERR "Restarting debugfs. Object ID: $objid\n" if $opt_v;
	$rcount = 0;
	print $dbgfs_h "quit\n";
	close $dbgfs_h;
	open $dbgfs_h, "|-", $cmd or die "*** Command $cmd failed: $!\n";
    }
    sleep $sec;
}
# cleanup 
print $dbgfs_h "quit\n";
close $dbgfs_h;

exit 0;
